<?php

/*
 * This Class polls a file generated by Systemd service 'sauce_serviced.service'
 * via Daemon.pm. That writes /usr/sausalito/sessions/.sauce_serviced_daemon.state 
 * with the JSON encoded data containing the keys 'file_count' (with the number of
 * services that have pending Systemd transactions),
 * The second key is 'event_files' and that contains an array of the names of the
 * services that are awaiting restarts.
 *
 * This Class reads the file if present. If not, it simulates an empty array of the
 * the expected keys and zero data for them. The result is then shown as JSON and
 * can be polled by other GUI classes to show the state of pending transactions.
 *
 * This Class has relaxed ACLs and any valid User can see the data. That's why it
 * is not included in /gui/metrics, which has stricter ACLs. 
 *
 * Note: This Class uses the cced-api and if that's down, it uses a fallback to the
 * old CCE.php Unix socket method.
 *
*/

namespace Gui\Controllers;
use App\Controllers\BaseController;
include_once("I18n.php");
require_once '/usr/sausalito/ui/chorizo/ci4/app/Helpers/blueonyx_helper.php';
use I18n;

class DaemonServices extends BaseController {

    public function __construct() {

    }

    public function index() {
        // This is hard-wired to use the API for now as first choice.
        // If the API doesn't respond, we fallback to CCE.php and the Unix socket:
        return $this->api_index();
    }

    public function api_index() {

        $referer = $_SERVER['HTTP_REFERER'] ?? '';
        if (!str_starts_with($referer, 'https://' . $_SERVER['HTTP_HOST'])) {
            // Access without an internal referer? Bye, bye:
            Log403Error("/gui/Forbidden403");
        }

        // Local call to cced-api (no auth needed from 127.0.0.1)
        $url = 'https://127.0.0.1:9092/v2/services';

        $ch = curl_init($url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);  // allow self-signed
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
        curl_setopt($ch, CURLOPT_FORBID_REUSE, false);    // allow reuse
        curl_setopt($ch, CURLOPT_FRESH_CONNECT, false);   // allow keep-alive

        $raw_data = curl_exec($ch);
        curl_close($ch);

        // Fallback logic
        if ($raw_data === false) {
            return $this->non_api_index();  // fallback on curl error
        }

        $data = json_decode($raw_data, true);
        if (json_last_error() !== JSON_ERROR_NONE || !is_array($data)) {
            // Fallback to Unix socket CCEd method if API is down:
            return $this->non_api_index();
        }

        http_response_code(200);
        $this->response->noCache();
        return $this->response->setJSON($data);
    }

    public function non_api_index() {

        $CI =& get_instance();
        $BX_SESSION = $CI->getBX_SESSION();

        // Only logged in users with 'serverShowActiveMonitor' get access to metrics:
        if (!$CI->getAllowed('validUser')) {
            // Nice people say goodbye, or CCEd waits forever:
            $CI->cceClient->bye();
            $CI->serverScriptHelper->destructor();
            Log403Error("/gui/Forbidden403");
        }

        $factory = $CI->serverScriptHelper->getHtmlComponentFactory("base-alpine", "/gui/metrics");
        $BxPage = $factory->getPage();
        $BxPage->setGETPOST(array('FORM_GET' => $this->request->getGet(), 'FORM_POST' => $this->request->getPost(), 'AGENT' => $this->request->getUserAgent()));
        $get_form_data = $BxPage->getGETPOST('GET');

        // Return ALL metrics:
        $query = '';
        $squery = '';

        // Or do we have a query string and return selected metrics only?
        if (isset($get_form_data['q'])) {
            $query = $get_form_data['q'];
            $squery = $query;
        }

        $filePath = '/usr/sausalito/sessions/.sauce_serviced_daemon.state';

        // Check if the file exists
        $have_json = 0;
        if (file_exists($filePath)) {
            // Read the file contents
            $jsonData = file_get_contents($filePath);

            // Decode the JSON data
            $data = json_decode($jsonData, true); // Use true to get an associative array

            // Check for JSON errors
            if (json_last_error() === JSON_ERROR_NONE) {
                // We have JSON data:
                $have_json = 1;
            }
        }

        if ($have_json === 0) {
            // If the JSON file wasn't there, we substitute blank content:
            $data = array('file_count' => 0, 'event_files' => array());
        }

        // Set Headers to no caching at all:
        $this->response->noCache();

        // Return metrics as JSON
        return $this->response->setJSON($data);
    }
}

/*
Copyright (c) 2008-2025 Michael Stauber, SOLARSPEED.NET
Copyright (c) 2008-2025 Team BlueOnyx, BLUEONYX.IT
All Rights Reserved.

1. Redistributions of source code must retain the above copyright 
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright 
   notice, this list of conditions and the following disclaimer in 
   the documentation and/or other materials provided with the 
   distribution.

3. Neither the name of the copyright holder nor the names of its 
   contributors may be used to endorse or promote products derived 
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.

You acknowledge that this software is not designed or intended for 
use in the design, construction, operation or maintenance of any 
nuclear facility.

*/
?>