<?php
/**
 * CceClient.php
 *
 * BlueOnyx CceClient for Codeigniter
 *
 * @package   CceClient
 * @author    Michael Stauber
 * @copyright Copyright (c) 2008-2025 Michael Stauber, SOLARSPEED.NET
 * @copyright Copyright (c) 2008-2025 Team BlueOnyx, BLUEONYX.IT
 * @copyright Copyright (c) 2003 Sun Microsystems, Inc.  All rights reserved.
 * @link      http://www.blueonyx.it
 * @license   http://devel.blueonyx.it/pub/BlueOnyx/licenses/SUN-modified-BSD-License.txt
 * @version   3.0
 */

//use App\Libraries\CceApiClient;

global $isCceClientDefined;
if ($isCceClientDefined) return;
$isCceClientDefined = true;

include_once ("CceError.php");
include_once ("System.php");
include_once ("CCE.php");
include_once ("CceApiClient.php");

class CceClient {
    
    public $handle;
    public $isConnected;
    public $NATIVE;
    public $Username;
    public $SessionId;
    public $Password;
    public $self;
    public $ERRORS = [];
    public $OID;
    public $cce_replay = [];
    public $cce_replay_file;
    public $replay_errors = [];
    public $CCE;

    private $cced_api_ip = '127.0.0.1';
    private $cced_api_port = '9092';
    private $apiClient;
    private $CI; // Cached CI instance

    //
    //-- Constructor:
    //

    public function __construct() {

        // Load in CI->get_instance() so that we don't have to call it a dozen times:
        $this->CI = & get_instance();

        if (is_file("/etc/DEBUG")) {
            $this->setDebug(true);
        }
        else {
            $this->setDebug(false);
        }

        $this->loadCceApiDefaultsFromConfig();

        // FALSE: Switch to API-only unless explicitly native fallback is used
        // TRUE:  Switch to CCE.php and use PHP and the Unix Socket

        // Default to native
        $this->setNative(TRUE);

        // Switch to API if it's NOT disabled manually AND reachable:
        if (!is_file('/etc/NOAPI')) {
            if ($this->isApiAvailable()) {
                $this->setNative(FALSE);
                if ($this->DEBUG) {
                    bx_error_log("CceClient::__construct(): API is available. Using API backend.");
                }
            }
            else {
                $this->setNative(TRUE);
                if ($this->DEBUG) {
                    bx_error_log("CceClient::__construct(): API unreachable or disabled. Falling back to native CCE backend.");
                }
            }
        }

        $this->CCE = new CCE();
        $this->handle = $this->CCE->ccephp_new();

        $this->isConnected = false;
    }

    function isApiAvailable() {
        try {
            $client = $this->getApiClient();
            $response = $client->ping();

            // Accept both 200 and 202 as valid
            return isset($response['status']) && in_array($response['status'], [200, 202]);
        }
        catch (Exception $e) {
            return false;
        }
    }

    // Set *different* API IP and port (we might want this for managing other BlueOnyx!):
    function setCceApiEndpoint($ip, $port) {
        $this->cced_api_ip = $ip;
        $this->cced_api_port = $port;
        $this->apiClient = new CceApiClient($ip, $port);
    }

    // Get API IP and port:
    function getCceApiEndpoint() {
        return [
            'ip' => $this->cced_api_ip,
            'port' => $this->cced_api_port
        ];
    }

    // Load API port from local cced-api config file:
    function loadCceApiDefaultsFromConfig() {
        $conf_file = '/etc/cced-api/config/cced-api.conf';
        if (is_readable($conf_file)) {
            $lines = file($conf_file, FILE_IGNORE_NEW_LINES | FILE_SKIP_EMPTY_LINES);
            foreach ($lines as $line) {
                if (preg_match('/^listen_address\s*=\s*(.*)$/', $line, $matches)) {
                    $parts = explode(':', $matches[1]);
                    if (count($parts) == 2) {
                        $this->cced_api_ip = $parts[0];
                        $this->cced_api_port = $parts[1];
                    }
                    break;
                }
            }
        }
        $this->apiClient = new CceApiClient($this->cced_api_ip, $this->cced_api_port);
    }

    function getApiClient() {
        if (!$this->apiClient) {
            $ep = $this->getCceApiEndpoint();
            $this->apiClient = new CceApiClient($ep['ip'], $ep['port']);
        }
        return $this->apiClient;
    }

    //
    //--- Legacy functions:
    //

    function setNative($CM) {
        $this->NATIVE = $CM;
    }

    function getNative() {
        return $this->NATIVE;
    }

    //
    //--- Generic class related functions:
    //

    function setUsername($Username = "") {
        $this->Username = $Username;
        $this->CI->setBX_SESSION_loginName($this->Username);
    }

    function getUsername() {
        if (!isset($this->Username)) {
            $BX_SESSION = $this->CI->getBX_SESSION();
            $this->Username = $BX_SESSION['loginName'];
        }
        return $this->Username;
    }

    function setSessionId($SessionId = "") {
        $this->SessionId = $SessionId;
        $this->CI->setBX_SESSION_sessionId($this->SessionId);
    }

    function getSessionId() {
        if (!isset($this->SessionId)) {
            $BX_SESSION = $this->CI->getBX_SESSION();
            $this->SessionId = $BX_SESSION['sessionId'];
        }
        return $this->SessionId;
    }

    function setPassword($Password = "") {
        $this->Password = $Password;
    }

    function getPassword() {
        return $this->Password;
    }

    function setDebug($DEBUG = false) {
        $this->DEBUG = $DEBUG;
    }

    function getDebug() {
        return $this->DEBUG;
    }

    function recordApiErrors($res, $oid = '', $key = '') {
        // Handle invalid response (no status)
        if (!isset($res['status'])) {
            $this->ERRORS[] = [
                'code' => 500,
                'oid' => $oid,
                'key' => $key,
                'message' => 'Invalid API response',
                'stack' => []
            ];
            return;
        }

        // Track recorded error codes to avoid duplicates
        $recordedCodes = [];

        // Check for warnings or errors in $res['data']['errors']
        if (isset($res['data']['errors']) && is_array($res['data']['errors'])) {
            foreach ($res['data']['errors'] as $err) {
                if (isset($err['code']) && $err['code'] >= 300 && isset($err['message'])) {
                    // Skip if this code has already been recorded
                    if (in_array($err['code'], $recordedCodes)) {
                        continue;
                    }

                    // Extract the message, removing "WARN" or "FAIL" prefix if present
                    $message = preg_replace('/^(WARN|FAIL)\s*/', '', $err['message']);
                    $this->ERRORS[] = [
                        'code' => $err['code'],
                        'oid' => $oid,
                        'key' => $key,
                        'message' => $message, // e.g., [[base-user.userNameAlreadyTaken]]
                        'stack' => []
                    ];

                    // Mark this code as recorded
                    $recordedCodes[] = $err['code'];
                }
            }
        }

        // Handle main error with status >= 300 (e.g., 401 FAIL)
        if ($res['status'] >= 300) {
            // Skip if this status code was already recorded from $res['data']['errors']
            if (in_array($res['status'], $recordedCodes)) {
                return;
            }

            $message = isset($res['message']) ? $res['message'] : 'Unknown error';
            $stack = isset($res['stack']) ? $res['stack'] : [];
            $this->ERRORS[] = [
                'code' => $res['status'],
                'oid' => $oid,
                'key' => $key,
                'message' => $message,
                'stack' => $stack
            ];
        }
    }

    //
    //--- CCEd related functions:
    //

    // description: authenticate to the server
    // param: userName: user name in string
    // param: password: password in string
    // returns: session ID if succeed, empty string if failed
    function auth($userName, $password) {
        if ($this->DEBUG) {
            bx_error_log("Command: AUTH \"$userName\" \"XXXX\"");
        }
        if (!$this->getNative()) {
            $result = $this->getApiClient()->auth($userName, $password);
            $this->recordApiErrors(['status' => empty($result) ? 401 : 200], '', 'auth');
            if (!empty($result)) {
                $this->setUsername($userName);
                $this->setSessionId($result);
                return $result;
            }
            return '';
        }
        return $this->CCE->ccephp_auth($userName, $password);
    }

    // description: authenticate using a session id
    function authkey($userName, $sessionId) {
        if ($this->DEBUG) {
            bx_error_log("Command: AUTHKEY \"$userName\" \"$sessionId\"");
        }
        if (!$this->getNative()) {
            $result = $this->getApiClient()->authkey($userName, $sessionId) ?? '';
            $this->recordApiErrors(['status' => empty($result) ? 401 : 200], '', 'authkey');
            return $result;
        }
        $this->CI->setBX_SESSION_sessionId($sessionId);
        return $this->CCE->ccephp_authkey($userName, $sessionId);
    }

    // description: determine the currently authenticated user
    // returns: the oid of the user object
    function whoami() {
        if ($this->DEBUG) {
            bx_error_log("Command: WHOAMI");
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->whoami($this->getUsername(), $this->getSessionId());
            $this->recordApiErrors($res, '', 'whoami');
            return $res['data']['oid'] ?? '';
        }
        return $this->CCE->ccephp_whoami();
    }

    // description: disconnect from cce
    function bye($whodidit = '') {
        if (($whodidit != '') && ($this->DEBUG)) {
            bx_error_log("CceClient: BYE initiated by $whodidit");
        }
        $this->isConnected = false;
        if ($this->DEBUG) {
            bx_error_log("Command: BYE");
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->bye($this->getUsername(), $this->getSessionId());
            $this->recordApiErrors($res);
            return in_array($res['status'], [200, 201]);
        }
        return $this->CCE->ccephp_bye();
    }

    // description: begin delayed-handler mode
    function begin() {
        if ($this->DEBUG) {
            bx_error_log("Command: BEGIN");
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->begin($this->getUsername(), $this->getSessionId());
            $this->recordApiErrors($res, '', 'begin');
            return isset($res['data']['oid']) ? $res['data']['oid'] : 0;
        }
        return $this->CCE->ccephp_begin();
    }

    // description: trigger all handlers to run since begin() call
    // returns: a success code based on the success or failure of all
    //          operations since begin()
    function commit() {
        if ($this->DEBUG) {
            bx_error_log("Command: COMMIT");
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->commit($this->getUsername(), $this->getSessionId());
            $this->recordApiErrors($res, '', 'commit');
            return isset($res['data']['oid']) ? $res['data']['oid'] : 0;
        }
        return $this->CCE->ccephp_commit();
    }

    // end authenticated session to cce
    function endkey() {
        if ($this->DEBUG) {
            bx_error_log("Command: ENDKEY");
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->endkey($this->getUsername(), $this->getSessionId());
            $this->recordApiErrors($res);
            return in_array($res['status'], [200, 201]);
        }
        else {
            return $this->CCE->ccephp_endkey();
        }
    }

    // description: create a CCE object of type $class, with properties in $vars
    // returns: oid of created object, or 0 on failure
    // usage: $oid = $cce->create($class, array( 'property' => 'value' ));
    function create($class, $vars = array()) {
        if ($this->DEBUG) {
            bx_error_log("Command: CREATE $class " . json_encode($vars));
        }

        // Prevent calling API with empty $vars
        if (empty($vars)) {
            bx_error_log("CceClient::create(): ABORTING! Empty \$vars for class $class.");
            return 0;
        }

        if (!$this->getNative()) {
            $res = $this->getApiClient()->create($this->getUsername(), $this->getSessionId(), $class, $vars);

            $this->recordApiErrors($res, '', $class);

            if (isset($res['status']) && in_array($res['status'], [200, 201])) {
                if (isset($res['data']['oid'])) {
                    return $res['data']['oid'];
                }
                if (isset($res['data']['oidlist'][0])) {
                    return $res['data']['oidlist'][0];
                }
            }

            return 0;
        }

        return $this->CCE->ccephp_create($class, $vars);
    }

    // description: destroy the CCE object with oid $oid
    // returns: boolean true for success, false for failure
    // usage: $ok = $cce->destroy($oid);
    function destroy($oid) {
        if ($this->DEBUG) {
            bx_error_log("Command: DESTROY $oid");
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->destroy($this->getUsername(), $this->getSessionId(), $oid);
            $this->recordApiErrors($res, $oid, '');
            return in_array($res['status'], [200, 201]);
        }
        return $this->CCE->ccephp_destroy($oid);
    }

    // description: destroy all objects of type $class that matching the
    // properties in $vars.
    // the match is done using 'find', so no findx parameters can be used
    // returns: nothing
    // usage: $cce->destroyObjects($class, array( 'findkey' => 'findvalue'));
    // DEPRECATED
    function destroyObjects($class, $vars = array()) {
        if ($this->DEBUG) {
            bx_error_log("Command: DestroyObjects $class " . json_encode($vars));
        }
        $oids = $this->find($class, $vars);
        foreach ($oids as $oid) {
            $this->destroy($oid);
        }
    }

    // the legacy find command
    // returns: matching $oids
    // usage: $oids = $cce->find($class, array( 'property' => 'value'));
    function find($class, $vars = array()) {
        if ($this->DEBUG) {
            bx_error_log("Command: FIND $class " . json_encode($vars));
        }
        if (!$this->getNative()) {

            $user = $this->getUsername() ?? '';
            $sid = $this->getSessionId() ?? '';

            if (empty($user) || empty($sid)) {
                $res = $this->getApiClient()->find(null, null, $class, $vars);
            }
            else {
                $res = $this->getApiClient()->find($user, $sid, $class, $vars);
            }

            $this->recordApiErrors($res, '', $class);
            return $res['data']['oidlist'] ?? array();
        }
        return $this->CCE->ccephp_find($class, $vars, "", 0);
    }

    // Description: advanced method of finding objects
    // $class : class to find
    // $vars : exact-match criteria
    // $revars : regex-match criteria
    // $sorttype : name of sorttype to use (optional)
    //           : listed in basetypes.schema, valid types are
    //           : ascii, old_numeric, locale, ip, hostname
    // $sortprop : name of property (key) on which to sort
    // returns: matching $oids
    // usage: $oids = $cce->findx($class, $vars, $regex_vars, $sorttype, $sortkey);
    function findx($class, $vars = array(), $revars = array(), $sorttype = "", $sortprop = "") {
        if ($this->DEBUG) {
            bx_error_log("Command: FINDX $class " . json_encode([
                'vars' => $vars,
                'revars' => $revars,
                'sorttype' => $sorttype,
                'sortprop' => $sortprop
            ]));
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->findx($this->getUsername(), $this->getSessionId(), $class, $vars, $revars, $sorttype, $sortprop);
            $this->recordApiErrors($res, '', $class);
            return $res['data']['oidlist'] ?? array();
        }
        return $this->CCE->ccephp_findx($class, $vars, $revars, $sorttype, $sortprop);
    }

    // description: gets the $namespace of the CCE object with given $oid
    // returns: a property hash or null if the object cannot be found
    // usage: $object = $cce->get($oid, $namespace);
    function get($oid, $namespace = "") {
        if ($this->DEBUG) {
            $msg = is_string($oid) ? "GET $oid" : "GET " . json_encode($oid);
            $msg .= ($namespace !== "") ? " . $namespace" : "";
            bx_error_log("Command: $msg");
        }

        if (!$this->getNative()) {
            $res = $this->getApiClient()->get($this->getUsername(), $this->getSessionId(), $oid, $namespace);

            // If GET failed (404, 401, 302, etc), return -1
            if (!isset($res['status']) || $res['status'] != 201 || !isset($res['data']['DATA'])) {
                if ($this->DEBUG) {
                    bx_error_log("CceClient::get(): API call failed or no DATA returned (status: " . ($res['status'] ?? 'null') . ")");
                }
                $this->recordApiErrors($res, $oid, $namespace);
                return -1;
            }

            $data = $res['data']['DATA'];

            // Normalize field names
            if (isset($data['oid'])) {
                $data['OID'] = $data['oid'];
                unset($data['oid']);
            }

            $data['NAMESPACE'] = $namespace;

            return $data;
        }

        return $this->CCE->ccephp_get($oid, $namespace);
    }

    // description: set object properties in CCE
    // returns: boolean true for success, boolean false for failure
    // usage: $ok = $cce->set($oid, $namespace, array( 'property' => 'value'));
    function set($oid, $namespace = "", $vars = array()) {
        if ($this->DEBUG) {
            $msg = "Command: SET $oid";
            if ($namespace != "") {
                $msg .= " $namespace";
            }
            $msg .= " " . json_encode($vars);
            bx_error_log($msg);
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->set($this->getUsername(), $this->getSessionId(), $oid, $namespace, $vars);

            // Store errors if they exist
            $this->recordApiErrors($res, $oid, $namespace);

            // Return "1" for success, "0" for failure, to match native behavior
            return in_array($res['status'], [200, 201]) ? 1 : 0;
        }
        return $this->CCE->ccephp_set($oid, $namespace, $vars);
    }

    // gets the names of the namespaces of an object or class
    // returns: array of strings, the names of the namespaces of the object
    // usage: $namespaces = $cce->names($oid); OR
    // usage: $namespaces = $cce->names($classname);
    function names($oidOrClass) {
        if ($this->DEBUG) {
            bx_error_log("Command: NAMES $oidOrClass");
        }

        if (!$this->getNative()) {
            $res = $this->getApiClient()->names(
                $this->getUsername(),
                $this->getSessionId(),
                $oidOrClass
            );

            // Return just the array, not the wrapper:
            $this->recordApiErrors($res, $oidOrClass, '');
            return $res['data']['namespaces'] ?? [];
        }

        return $this->CCE->ccephp_names($oidOrClass);
    }

    // description: determine if CCE is suspended or not
    // returns: reason string if suspended, false otherwise
    function suspended() {
        if ($this->DEBUG) {
            bx_error_log("Command: SUSPENDED");
        }
        if (!$this->getNative()) {
            $res = $this->getApiClient()->suspended($this->getUsername(), $this->getSessionId());
            $this->recordApiErrors($res);
            return $res['data'] ?? false;
        }
        return $this->CCE->ccephp_suspended();
    }

    //
    //--- Done until here
    //

    // description: set CCE read-only.
    // Requires: systemAdministrator access
    // param: reason: reason for CCE being read-only
    // returns: true is successful
    function suspend($reason) {
        if ($this->DEBUG) {
            bx_error_log("Command: SUSPEND \"$reason\"");
        }

        if (!$this->getNative()) {
            $res = $this->getApiClient()->suspend($this->getUsername(), $this->getSessionId(), $reason);
            $this->recordApiErrors($res);
            return isset($res['status']) && in_array($res['status'], [200, 201]);
        }

        return $this->CCE->ccephp_suspend($reason);
    }

    // description: set CCE read-write after a call to suspend().
    // Requires: systemAdministrator access
    // returns: true is successful
    function resume() {
        if ($this->DEBUG) {
            bx_error_log("Command: RESUME");
        }

        if (!$this->getNative()) {
            $res = $this->getApiClient()->resume($this->getUsername(), $this->getSessionId());
            $this->recordApiErrors($res);
            return isset($res['status']) && in_array($res['status'], [200, 201]);
        }

        return $this->CCE->ccephp_resume();
    }

    //
    //--- CCE Replay:
    //
    // This works similar to 'Delayed Transactions', but uses a different method.
    // You can store SET transactions in a JSON encoded replay file. Several
    // independent transactions can be stored. First Array key is the number of the
    // transaction. The value of each transaction is an Array as well. Index being
    // the OID that the SET transaction is performed against and the rest of that
    // Array being the key/value pairs to be performed.
    //
    // description: begin of CCE-Replay transaction (record to file):
    function record($oid, $namespace = "", $vars = array()) {
        if ($this->DEBUG) {
            bx_error_log("Command: RECORD REPLAY");
        }

        // Setup error array:
        //$this->replay_errors = array();
        // Get Username and SessionID:
        $uname = $this->getUsername();
        $sessionId = $this->getSessionId();

        // Load file-helper class:
        $CI = & get_instance();

        // Define target file coupled to this uname:
        $this->cce_replay_file = '/usr/sausalito/license/json_cce_replay.' . $uname;

        // Open replay file if it exists:
        if (is_file($this->cce_replay_file)) {
            $json_cce_replay = file_get_contents($this->cce_replay_file);
            $this->cce_replay = json_decode($json_cce_replay, true);
        }

        if (!is_array($this->cce_replay)) {
            // Array to store CCEd transactions for later replay:
            $this->cce_replay = array();
        }

        // Append Replay Transaction:
        $this->cce_replay[][$oid] = array(
            'namespace' => $namespace,
            'transaction' => $vars
        );

        // Encode Replay Transaction:
        $json_cce_replay = json_encode($this->cce_replay);

        // Delete old replay file:
        $nfk = '';
        $ret = $CI->serverScriptHelper->shell("/bin/rm -f " . $this->cce_replay_file, $nfk, 'root', $sessionId);

        // Write the new license file out to disk:
        if (!write_file($this->cce_replay_file, $json_cce_replay)) {
            bx_error_log("ERROR: " . 'Could not write CCE replay-file ' . $this->cce_replay_file);
            $this->setReplayError('2', $oid, $this->cce_replay_file, 'Could not write CCE replay-file ' . $this->cce_replay_file);
            return false;
        }
        else {
            // No errors:
            return true;
        }
    }

    // description: Deletes an existing CCE-Replay file without executing it:
    function replayReset() {
        // Load file-helper class:
        $CI = & get_instance();

        // Get SessionID:
        $uname = $this->getUsername();
        $sessionId = $this->getSessionId();

        // Define target file coupled to this uname:
        $this->cce_replay_file = '/usr/sausalito/license/json_cce_replay.' . $uname;

        if (is_file($this->cce_replay_file)) {
            // Delete old replay file:
            bx_error_log("Deleting " . $this->cce_replay_file);
            $ret = $CI->serverScriptHelper->shell("/bin/rm -f " . $this->cce_replay_file, $nfk, 'root', $sessionId);
            if (is_file($this->cce_replay_file)) {
                bx_error_log("Shits not gone! " . $this->cce_replay_file);
            }
        }
        return true;
    }

    // description: Report how many stored transactions (if any) are in the current CCE-Replay file.
    function replayStatus() {
        // Load file-helper class:
        $CI = & get_instance();

        // Get SessionID:
        $uname = $this->getUsername();
        $sessionId = $this->getSessionId();

        // Define target file coupled to this uname:
        $this->cce_replay_file = '/usr/sausalito/license/json_cce_replay.' . $uname;

        if (is_file($this->cce_replay_file)) {
            $json_cce_replay = file_get_contents($this->cce_replay_file);
            $this->cce_replay = json_decode($json_cce_replay, true);

            // Check if it contains an Array:
            $num_of_stored_transactions = '0';
            if (is_array($this->cce_replay)) {
                $num_of_stored_transactions = count(array_keys($this->cce_replay));
            }
            return $num_of_stored_transactions;
        }
        else {
            // No replay file present:
            return '-1';
        }
    }

    // description: Execution of transactions stored in the CCE-Replay file.
    function replay($replayType = "auto") {

        // Load file-helper class:
        $CI = & get_instance();

        // Get SessionID:
        $uname = $this->getUsername();
        $sessionId = $this->getSessionId();

        // Define target file coupled to this uname:
        $this->cce_replay_file = '/usr/sausalito/license/json_cce_replay.' . $uname;

        if (is_file($this->cce_replay_file)) {
            $json_cce_replay = file_get_contents($this->cce_replay_file);
            $this->cce_replay = json_decode($json_cce_replay, true);

            // Check if it contains an Array:
            $num_of_stored_transactions = '0';
            if (is_array($this->cce_replay)) {
                $num_of_stored_transactions = count(array_keys($this->cce_replay));
                if (($replayType == "auto") && ($num_of_stored_transactions > '0')) {
                    foreach ($this->cce_replay as $replayKey => $ReplayTransAction) {
                        foreach ($ReplayTransAction as $OID => $SetVal) {
                            $this->set($OID, $SetVal['namespace'], $SetVal['transaction']);
                            $num_of_stored_transactions--;
                            $this->replay_errors = $this->errors();
                            if (count($this->replay_errors) > '0') {
                                // We had an error. Assume the replay-file is messed up. Delete it:
                                $ret = $CI->serverScriptHelper->shell("/bin/rm -f " . $this->cce_replay_file, $nfk, 'root', $sessionId);
                                bx_error_log("ERROR: " . 'Deleting messed up CCE Replay-File (' . $this->cce_replay_file . ').');
                                return $this->replay_errors;
                            }
                        }
                    }
                }
                else {
                    // Take first transaction out of the Array and execute just this single transaction:
                    $one_step = array_shift($this->cce_replay);
                    foreach ($one_step as $OID => $SetVal) {
                        $CI->cceClient->set($OID, $SetVal['namespace'], $SetVal['transaction']);
                        $this->replay_errors = $CI->cceClient->errors();
                        if (count($this->replay_errors) > '0') {
                            // We had an error. Assume the replay-file is messed up. Delete it:
                            $ret = $CI->serverScriptHelper->shell("/bin/rm -f " . $this->cce_replay_file, $nfk, 'root', $sessionId);
                            bx_error_log("ERROR: " . 'Deleting messed up CCE Replay-File (' . $this->cce_replay_file . ').');
                            return $this->replay_errors;
                        }
                    }
                    $num_of_stored_transactions = count($this->cce_replay);
                    if ($num_of_stored_transactions != '0') {
                        // Write rest of unprocessed transactions back to file:
                        $json_cce_replay = json_encode($this->cce_replay);
                        write_file($this->cce_replay_file, $json_cce_replay);
                        bx_error_log("INFO: " . "Updating up CCE Replay-File with $num_of_stored_transactions remaining transactions (" . $this->cce_replay_file . ').');
                        return $num_of_stored_transactions;
                    }
                }
            }
            else {
                bx_error_log("ERROR: " . 'CCE Replay-File (' . $this->cce_replay_file . ') did not contain a readable Array of stored transactions.');
                $this->setReplayError('2', '', $this->cce_replay_file, 'CCE Replay-File (' . $this->cce_replay_file . ') did not contain a readable Array of stored transactions.');
                return false;
            }

            // Delete replay file:
            $nfk = '';
            $ret = $CI->serverScriptHelper->shell("/bin/rm -f " . $this->cce_replay_file, $nfk, 'root', $sessionId);

            return $num_of_stored_transactions;
        }
    }

    // description: Replay transactions cannot use regular CCEd error messages, so we mimick them with this function:
    function setReplayError($code, $oid = "", $key = "", $msg = "") {
        $numErrs = count($this->ERRORS);
        $this->ERRORS[$numErrs]['code'] = $code;
        $this->ERRORS[$numErrs]['oid'] = $oid;
        $this->ERRORS[$numErrs]['key'] = $key;
        $this->ERRORS[$numErrs]['message'] = $msg;
        if ($numErrs != "0") {
            $numErrs++;
        }
    }

    // description: Function to poll replay errors:
    function get_replay_errors() {
        return $this->ERRORS;
    }

    // description: connect to CCE
    // param: socketPath: the path of the Unix domain socket to CCE
    // returns: true if succeed, false otherwise
    function connect($socketPath = "") {
        if ($socketPath == "") {
            // get from config file
            $system = new System();
            $socketPath = $system->getConfig("ccedSocketPath");
        }

        if (!$this->getNative()) {
            // API-based backend doesn't need a socket connection;
            // just mark as "connected" if API endpoint is reachable
            $res = $this->getApiClient()->ping();
            if ((isset($res['message'])) && ($res['message'] === 'PONG')) {
                $this->isConnected = true;
            }
            else {
                $this->isConnected = false;
            }
        }
        else {
            $this->isConnected = $this->CCE->ccephp_connect($socketPath);
        }

        return $this->isConnected;
    }

    // description: get the last occured error
    // returns: an array of CceError objects
    // usage: $errors = $cce->errors();
    function errors() {
        $errorObjs = array();

        if (!$this->getNative()) {
            $errors = $this->ERRORS;
        }
        else {
            $errors = $this->CCE->ccephp_errors();
        }

        // Merge replay errors if present
        if (is_array($this->get_replay_errors()) && count($this->get_replay_errors()) > 0) {
            $errors = array_merge($errors, $this->get_replay_errors());
        }

        foreach ($errors as $error) {
            $ekey = "";
            $msg = "";

            // Determine 'key'
            if (isset($error["key"])) {
                if (is_array($error["key"])) {
                    $ekey = reset($error["key"]);
                    if ($this->DEBUG) {
                        bx_error_log("CCE Error: 'key' was array: " . json_encode($error["key"]));
                    }
                }
                else {
                    $ekey = $error["key"];
                }
            }

            // Determine 'message'
            if (!empty($error["message"])) {
                $msg = $error["message"];
            }
            elseif (!empty($error["code"])) {
                $msg = "Error code " . $error["code"];
            }
            elseif (!empty($ekey)) {
                $msg = "Error in key: " . $ekey;
            }
            else {
                $msg = "Unknown error";
            }

            $errorObjs[] = new CceError(
                $error["code"] ?? "UNKNOWN",
                $error["oid"] ?? "-1",
                $ekey,
                $msg,
                [
                    "code" => $error["code"] ?? "UNKNOWN",
                    "oid" => $error["oid"] ?? "-1",
                    "key" => $ekey
                ]
            );
        }

        // Also include any replay_errors manually injected
        if (!empty($this->replay_errors)) {
            foreach ($this->replay_errors as $value) {
                $errorObjs[] = $value;
            }
        }

        return $errorObjs;
    }

    // description: returns array of hashes.  Each hash contains
    // information about a particular error, include "code", "oid",
    // "key", and "message".
    function raw_errors() {
        if (!$this->getNative()) {
            return ccephp_errors($this->handle);
        }
        else {
            return $this->CCE->ccephp_errors();
        }
    }

    // find objects, returning a list sorted by the values in
    // the $key attribute
    // sort is ascii sort
    // returns: matching $oids
    // usage: $oids = $cce->findSorted($class, $sortkey, array( 'property' => 'value'));
    function findSorted($class, $key, $vars = array()) {
        if ($this->DEBUG) {
            bx_error_log("Command: FINDSORTED $class $key " . json_encode($vars));
        }

        if (!$this->getNative()) {
            // Convert all boolean values to "1"/"0" strings for API compatibility
            $clean_vars = array_map(function ($val) {
                if (is_bool($val)) {
                    return $val ? "1" : "0";
                }
                return (string)$val;
            }, $vars);

            $res = $this->getApiClient()->findx(
                $this->getUsername(),
                $this->getSessionId(),
                $class,
                $clean_vars,
                [],         // no regex args
                "ascii",    // sort type
                $key        // sort key
            );

            $this->recordApiErrors($res, '', $class);
            return $res['data']['oidlist'] ?? [];
        }

        return $this->CCE->ccephp_find($class, $vars, $key, 0);
    }

    // like findSorted, but sort is numeric instead of ascii.
    // returns: matching $oids
    // usage: $oids = $cce->findSorted($class, $sortkey, array( 'property' => 'value'));
    function findNSorted($class, $key, $vars = array()) {
        if ($this->DEBUG) {
            bx_error_log("Command: FINDNSORTED $class $key " . json_encode($vars));
        }

        if (!$this->getNative()) {
            // Ensure all values in $vars are strings (CCE requires string values)
            $clean_vars = array_map(function ($val) {
                if (is_bool($val)) {
                    return $val ? "1" : "0";
                }
                return (string)$val;
            }, $vars);

            $res = $this->getApiClient()->findx(
                $this->getUsername(),
                $this->getSessionId(),
                $class,
                $clean_vars,
                [],           // no regex args
                "numeric",    // sort type
                $key          // sort key
            );

            $this->recordApiErrors($res, '', $class);
            return $res['data']['oidlist'] ?? [];
        }

        return $this->CCE->ccephp_find($class, $vars, $key, 1);
    }

    // description: get a CCE object matching given properties
    //              the match is done using 'find', so no findx parameters
    //              can be used
    // returns: a property hash or null if the object cannot be found
    // usage: $object = $cce->getObject($class, array( 'property' => 'value'), $namespace);
    function getObject($class, $vars = array(), $namespace = "") {
        $use_cache = '1';

        // Check if $vars 'cce_nocache' is present and set to 'cce_nocache'
        if (is_array($vars) && in_array('cce_nocache', $vars)) {
            $use_cache = '0';
            unset($vars['cce_nocache']);
            if ($this->DEBUG) {
                bx_error_log("Command: GetObject with vars 'cce_nocache' set for $class " . json_encode($vars) . ' ' . $namespace);
            }
        }

        if ($this->DEBUG) {
            bx_error_log("Command: GetObject $class " . json_encode($vars) . ' ' . $namespace);
        }

        $cacheworthy_oids = array('System', 'ActiveMonitor', 'Capabilities', 'MySQL', 'PHP');

        $oids = [];

        if ((in_array($class, $cacheworthy_oids)) && (count($vars) == 0) && ($use_cache == '1')) {
            if ($class == 'System' && $namespace == '' && isset($_SESSION['System']) && is_array($_SESSION['System'])) {
                if ($this->DEBUG) {
                    bx_error_log("CceClient.getObject(): Using Session saved 'System' object");
                }
                return $_SESSION['System'];
            }

            if (isset($_SESSION['OIDS'][$class])) {
                $oids = $_SESSION['OIDS'][$class];
                if ($this->DEBUG) {
                    bx_error_log("CceClient.getObject(): Using Session saved OID for $class");
                }
            }
            else {
                if (!$this->getNative()) {
                    $res = $this->getApiClient()->find($this->getUsername(), $this->getSessionId(), $class, $vars);
                    $this->recordApiErrors($res, '', $class);
                    $oids = $res['data']['oidlist'] ?? [];
                }
                else {
                    $oids = $this->find($class, $vars);
                }

                $data = isset($_SESSION['OIDS']) ? ['OIDS' => array_merge($_SESSION['OIDS'], [$class => $oids])] : ['OIDS' => [$class => $oids]];
                if ($this->DEBUG) {
                    bx_error_log("CceClient.getObject(): Storing OID for $class in session-data");
                }
                session()->set($data);
            }
        }
        else {
            if (!$this->getNative()) {
                $res = $this->getApiClient()->find($this->getUsername(), $this->getSessionId(), $class, $vars);
                $this->recordApiErrors($res, '', $class);
                $oids = $res['data']['oidlist'] ?? [];
            }
            else {
                $oids = $this->find($class, $vars);
            }
        }

        // We have OIDs, so we run a GET and return the Object
        if (isset($oids) && count($oids) > 0 && isset($oids[0])) {
            if ($class == 'System' && $namespace == '') {
                $SystemObj = $this->get($oids[0], $namespace);
                if ($this->DEBUG) {
                    bx_error_log("CceClient.getObject(): Caching System object in Session-Data");
                }
                session()->set(['System' => $SystemObj]);
                return $SystemObj;
            }
            else {
                return $this->get($oids[0], $namespace);
            }
        }

        return NULL;
    }

    // description: get an array of CCE objects matching given properties
    //              the match is done using 'find', so no findx parameters
    //              can be used
    //
    // returns:     an array of property hashes, or null if no objects can be found
    //
    // usage:       $objects = $cce->getObjects($class, array( 'property' => 'value'), $namespace);
    function getObjects($class, $vars = [], $namespace = "") {
        if ($this->DEBUG) {
            bx_error_log("Command: GetObjects $class " . json_encode($vars) . " " . $namespace);
        }
        $oids = $this->find($class, $vars);
        return array_map(fn($oid) => $this->get($oid, $namespace), $oids);
    }

    // cced-api has a GETALL function to fetch ALL Objects WITH their NameSpaces
    // And for usage with CCE.php we can fall back to /usr/sausalito/sbin/external_cce_get.pl
    function getAll($class_or_oids, $args = []) {
        if (!$this->getNative()) {
            $res = $this->getApiClient()->getAll($this->getUsername(), $this->getSessionId(), $class_or_oids, $args);
            // This ceates an "Illegal API response" error during Vsite create:
            //$this->recordApiErrors($res, '', 'getAll');
            return $res;
        }
        // Native fallback via /usr/sausalito/sbin/external_cce_get.pl:
        $oids = $this->find($class_or_oids, $args);
        $oid_list = json_encode($oids);
        $output = '';
        $ret = $this->CI->serverScriptHelper->shell("/usr/sausalito/sbin/external_cce_get.pl --oid $oid_list", $output, 'root', $this->getSessionId());

        // Decode results:
        $JSON_OBJECTS = json_decode($output, true);
        $data = $JSON_OBJECTS;

        if ($ret != 0 || !is_array($JSON_OBJECTS)) {
            if ($this->DEBUG) {
                bx_error_log("getAll(): external_cce_get.pl failed. Return: $ret, Output: $output");
            }
            return '-1';
        }

        if (is_array($data)) {
            return $data;
        }
        // Failed!
        return '-1';
    }

    // returns: true if the client is connected to the server, false otherwise
    function isConnected() {
        return $this->isConnected;
    }

    // description: set a CCE object
    //              the match is done using 'find', so no findx parameters
    //              can be used
    // returns: true on success, false otherwise
    // usage: $ok = $cce->set($class, array( 'property' => 'value'), $namespace, array( 'findkey' => 'findvalue'));
    function setObject($class, $setVars = array() , $namespace = "", $findVars = array()) {
        if ($this->DEBUG) {
            bx_error_log("Command: setObject $class " . json_encode($setVars) . ' ' . $namespace . ' ' . json_encode($findVars));
        }
        $oids = $this->find($class, $findVars);
        if (count($oids) > 0) {
            return $this->set($oids[0], $namespace, $setVars);
        }
        else {
            return 0;
        }
    }

    // description: set a CCE object. If it does not exist, create it
    //              the match is done using 'find', so no findx parameters
    //              can be used
    // returns: true on success, false otherwise
    // usage: $ok = $cce->set($class, array( 'property' => 'value'), $namespace, array( 'findkey' => 'findvalue'));
    function setObjectForce($class, $setVars = array() , $namespace = "", $findVars = array()) {
        if ($this->DEBUG) {
            bx_error_log("Command: setObjectForce $class " . json_encode($setVars) . ' ' . $namespace . ' ' . json_encode($findVars));
        }
        $oids = $this->find($class, $findVars);
        // create if necessary
        if (count($oids) == 0) {
            if (!$this->create($class, $setVars)) return 0;
        }

        return $this->setObject($class, $setVars, $namespace, $findVars);
    }

    // description: determines if the current session is a handler in rollback mode.
    function isRollback() {
        if (!$this->getNative()) {
            return ccephp_is_rollback($this->handle);
        }
        else {
            return $this->CCE->ccephp_is_rollback();
        }
    }

    // description: converts a array into a CCE-encoded scalar
    function array_to_scalar($array) {
        $result = "&";
        if (is_array($array)) {
            $result = "&";
            foreach ($array as $value) {
                $value = preg_replace_callback("/([^A-Za-z0-9_\. -])/", function ($m) {
                    return sprintf('%%%02X', ord($m[1]));
                }
                , $value);
                $value = preg_replace("/ /", "+", $value);

                $result .= $value . "&";
            }
        }
        if ($result == "&") $result = "";
        return $result;
    }

    // description: converts a CCE-encoded scalar into an array
    function scalar_to_array($scalar) {

        // just in case trim off whitespace
        $scalar = trim($scalar);
        $scalar = preg_replace("/^&/", "", $scalar);
        $scalar = preg_replace("/&$/", "", $scalar);
        $scalar = preg_replace("/;/", "", $scalar);

        $array = explode("&", $scalar);

        for ($i = 0;$i < count($array);$i++) {
            $array[$i] = preg_replace("/\+/", " ", $array[$i]);
            $array[$i] = preg_replace_callback("/%([0-9a-fA-F]{2})/", function ($m) {
                return chr(hexdec($m[1]));
            }
            , $array[$i]);
        }
        $array = array_filter($array);
        return $array;
    }

    // description: converts a string to a CCE-encoded scalar.
    // This is new as of 5200R and is a necessity due to CodeIgniters
    // XSS cleaning of our form data:
    function string_to_scalar($string) {
        // Just in case trim off whitespace:
        $string = trim($string);

        // Strip leading and trailing "&" - just in case as well:
        $string = preg_replace("/^&/", "", $string);
        $string = preg_replace("/&$/", "", $string);

        // Strip excess whitespaces:
        $string = preg_replace("/\s\s+/", " ", $string);

        // Replace ", " with "&":
        $string = preg_replace("/,[\s+]{0,999}/i", "&", $string);

        // Replace "\n" with "&":
        $string = preg_replace("/\n/i", "&", $string);

        // Build scalar:
        if ($string) {
            $scalar = "&" . $string . "&";
        }
        else {
            $scalar = "";
        }

        return $scalar;
    }

    // description: converts a CCE-encoded scalar into a string:
    function scalar_to_string($scalar) {
        if (preg_match("/^\&(.*)\&$/", $scalar, $regs)) {
            $value = implode("\n", $this->scalar_to_array($scalar));
        }
        else {
            $value = $scalar;
        }
        return $value;
    }

    // _escape: This function is used to clean up data coming
    // from the GUI in a fashion so that it can be stored into CODB.
    function _escape($text) {
        if (is_array($text)) {
            // We have an array. Transform it into a scalar for easier processing:
            $text = array_to_scalar($text);
        }

        // Check if this is a simple matter. If so, return right away:
        if (preg_match('/^[a-zA-Z0-9_]+$/', $text)) {
            return $text;
        }

        // Replace unwanted chars with their double escaped counterparts or another safe replacement:
        $out = str_replace(array(
            "\\",
            "\a",
            "\b",
            "\f",
            "\n",
            "\t",
            '"',
            '$',
            '&quot;',
            '&amp;',
            '&#39;',
            '&lt;',
            '&gt;'
        ) , array(
            "\\\\",
            "\\a",
            "\\b",
            "\\f",
            "\\n",
            "\\t",
            "\\\"",
            "\\$",
            '\"',
            '\&',
            "'",
            '<',
            '>'
        ) , $text);
        return $out;
    }

}

/*
Copyright (c) 2008-2025 Michael Stauber, SOLARSPEED.NET
Copyright (c) 2008-2025 Team BlueOnyx, BLUEONYX.IT
Copyright (c) 2003 Sun Microsystems, Inc. 
All Rights Reserved.

1. Redistributions of source code must retain the above copyright 
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright 
   notice, this list of conditions and the following disclaimer in 
   the documentation and/or other materials provided with the 
   distribution.

3. Neither the name of the copyright holder nor the names of its 
   contributors may be used to endorse or promote products derived 
   from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.

You acknowledge that this software is not designed or intended for 
use in the design, construction, operation or maintenance of any 
nuclear facility.

*/
?>